program MSVCRTTest;

{$APPTYPE CONSOLE}

uses
  SysUtils,
  Math,
  Windows,
  MSVCRT in 'MSVCRT.pas',
  Console in 'Console.pas';

// Console unit can be found on my website:
  // http://rvelthuis.de/programs/console.html

{$LINK 'msvcrttest.obj'}

procedure _mathtest; cdecl; external;
procedure _stdlibtest; cdecl; external;
procedure _stdiotest; cdecl; external;
procedure _timetest; cdecl; external;
procedure _ctypetest; cdecl; external;
procedure _stringtest; cdecl; external;
procedure _wchartest; cdecl; external;

type
  MenuItem = record
    Id: Integer;
    Str: string;
  end;

const
  demoNone       = 0;
  demoMath       = 1;
  demoStdlib     = 2;
  demoStdio      = 3;
  demoTime       = 4;
  demoCtype      = 5;
  demoString     = 6;
  demoWchar      = 7;
  demoExit       = 255;

  MenuItems: array[0..7] of MenuItem = (
    (Id: demoMath; Str: 'math.h demo'),
    (Id: demoStdlib; Str: 'stdlib.h demo'),
    (Id: demoStdio; Str: 'stdio.h demo'),
    (Id: demoTime; Str: 'time.h demo'),
    (Id: demoCtype; Str: 'ctype.h demo'),
    (Id: demoString; Str: 'string.h demo'),

    // Enter new demos here.

    (Id: demoExit; Str: 'Exit MSVCRTTest'),
    (Id: demoNone; Str: 'You should never see this'));

procedure Center(S: string);
begin
  GotoXY((80 - Length(S)) div 2, WhereY);
  Write(S);
end;

function Menu: Integer;
var
  I, Max, Num: Integer;
  C: Char;
begin
  TextColor(Yellow);
  Center('Demo and test program for MSVCRT.pas');
  TextColor(White);

  // Count items and find longest string.
  Num := 0;
  Max := 0;
  while MenuItems[Num].Id <> demoNone do
  begin
    if Max < Length(MenuItems[Num].Str) then
      Max := Length(MenuItems[Num].Str);
    Inc(Num);
  end;

  // Write menu items.
  for I := 0 to Num - 1 do
  begin
    GotoXY(1, I * 2 + 4);
    Center(Format('%2d: %-*s', [I + 1, Max, MenuItems[I].Str]));
  end;

  // Prompt for input.
  GotoXY(1, Num * 2 + 5);
  TextColor(Yellow);
  Center(Format('Please enter your choice [1..%d]: ', [Num]));
  TextColor(White);

  // Read valid key.
  repeat
    C := ReadKey;
    if C = #27 then
      I := Num - 1
    else
      I := Ord(C) - Ord('1');
  until I in [0..Num];

  // Return matching Id.
  Result := MenuItems[I].Id;
end;

// prompt for keypress.
procedure Pause;
begin
  Writeln;
  TextColor(Yellow);
  Write('Press any key... ');
  ReadKey;
  TextColor(White);
  Writeln;
  Writeln;
end;

var
  MenuId: Integer;

begin
  // Loop menu / selected demo.
  TextBackground(Black);
  TextColor(White);
  repeat
    ClrScr;
    MenuId := Menu;
    ClrScr;
    case MenuId of
      demoMath:
        _mathtest;
      demoStdlib:
        _stdlibtest;
      demoStdio:
        _stdiotest;
      demoTime:
        _timetest;
      demoCtype:
        _ctypetest;
      demoString:
        _stringtest;
      demoExit:
        Exit;
    end;
    Pause;
  until False;
end.
