/*
   File name:  MsvcrtTest.c
   Function:   Testng the functions implemented in MSVCRT.pas
   Language:   Borland C
   Author:     Rudy Velthuis
   Copyright:  (c)2006 Rudy Velthuis
*/


#pragma hdrstop

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <string.h>
#include <locale.h>

#include <windows.h>
#include <conio.h>

#include "msvcrttest.h"

// New line.
static void nl(void)
{
    puts("\n");
}

// Display string underlined.
static void caption(const char *s)
{
    char buffer[256];
    memset(buffer, 0, sizeof(buffer));
    memset(buffer, '-', strlen(s));
    printf("\n%s\n%s\n\n", s, buffer);
}

// Print double formatted
static void printd(const char *s, double d)
{
    char buffer[256];

    sprintf(buffer, "%s:", s);
    printf("%-30s %12.6f\n", buffer, d);
}

// Print long int formatted
static void printl(const char *s, long i)
{
    char buffer[256];

    sprintf(buffer, "%s:", s);
    printf("%-30s %12ld\n", buffer, i);
}
// Calls abs, acos, asin, atan, atan2, ceil, cos, cosh, exp, fabs,
// floor, fmod, frexp, labs, ldexp, log, log10, modf, pow, sin, sinh
// sqrt, tan, tanh

void mathtest(void)
{
    double mantissa;
    int exponent;

    double fractional;
    double integral;

    caption("math.h");

//- abs(), acos(), asin(), atan(), atan2(), ceil(), cos(), sin(), cosh()
//- exp(), fabs(), floor(), fmod(), labs(), log(), log10(), pow(), sinh()
//- sqrt() tanh()

    printl("abs(-1234)", abs(-1234));
    printd("acos(cos(1.1))", acos(cos(1.1)));
    printd("asin(sin(1.1))", asin(sin(1.1)));
    printd("atan(tan(1.1))", atan(tan(1.1)));
    printd("atan2(2.0 * tan(1.1), 2.0)", atan2(2.0 * tan(1.1), 2.0));
    printd("ceil(7.92)", ceil(7.92));
    printd("cos(acos(0.5))", cos(acos(0.5)));
    printd("sin(asin(0.5))", sin(asin(0.5)));
    printd("cosh(2.0)", cosh(2.0));
    printd("exp(log(123))", exp(log(123)));
    printd("fabs(-123.456)", fabs(-123.456));
    printd("floor(7.92)", floor(7.92));
    printd("fmod(3.6, 2.4)", fmod(3.6, 2.4));
    printd("fmod(-3.6, 2.4)", fmod(-3.6, 2.4));
    printl("labs(-12345L)", labs(-12345L));
    printd("log(exp(17))", log(exp(17)));
    printd("log10(1.0e19)", log10(1.0e19));
    printd("pow(0.64, 0.5)", pow(0.64, 0.5));
    printd("sinh(2.0)", sinh(2.0));
    printd("sqrt(1.44)", sqrt(1.44));
    printd("tanh(2.0)", tanh(2.0));

//- frexp(), ldexp()

    mantissa = frexp(0.75 * 128, &exponent);
    printd("frexp(0.75 * 128) mantissa", mantissa);
    printl("frexp(0.75 * 128) exponent", exponent);
    printd("ldexp(mantissa, exponent)", ldexp(mantissa, exponent));

//- modf()

    fractional = modf(9.8765, &integral);
    printd("modf(9.8765) integral", integral);
    printd("modf(9.8765) fractional", fractional);
}

int cmpint(const void *i1, const void *i2)
{
    if (*(int *)i1 > *(int *)i2)
        return 1;
    else if (*(int *)i1 < *(int *)i2)
        return -1;
    else
        return 0;
}

// Calls: abort, atof, atoi, atol, atexit, bsearch, calloc, exit, free,
// getenv, malloc, mblen, mbstowcs, mbtowc, qsort, rand, realloc, srand,
// system, wctomb, strtol, strtoul, strtod,

void stdlibtest(void)
{
    int i;
    int *ptr;
    int array[10];
    char *endptr;
    double d;
    char ansi[256];
    wchar_t wide[256];
    long int l;
    unsigned long int ul;
    div_t dv;
    ldiv_t ldv;

    int error = 0;

    caption("stdlib.h");

//- div(), ldiv()

    dv = div(100, 17);
    printf("100 div 17 = %d remainder %d\n", dv.quot, dv.rem);
    ldv = ldiv(100000L, 3000L);
    printf("100000L div 3000L = %ld remainder %d\n", ldv.quot, ldv.rem);
    nl();


//- atof(), atol(), atoi()

    printf("atof(\"123.456\"): %f\n", atof("123.456"));
    printf("atof(\"1234567\"): %f\n", atof("1234567"));
    printf("atof(\"1234abc\"): %f\n", atof("1234abc"));
    nl();

    printf("atol(\"123.456\"): %ld\n", atol("123.456"));
    printf("atol(\"1234567\"): %ld\n", atol("1234567"));
    printf("atol(\"1234abc\"): %ld\n", atol("1234abc"));
    nl();

    printf("atoi(\"123.456\"): %d\n", atoi("123.456"));
    printf("atoi(\"1234567\"): %d\n", atoi("1234567"));
    printf("atoi(\"1234abc\"): %d\n", atoi("1234abc"));
    nl();

//- strtol(), strtoul(), strtod()

    l = strtol("-123abz", &endptr, 10);
    printf("strtol(\"-123abz\", &endptr, 10):  %ld  %08X  (\"%s\")\n",
        l, l, endptr);
    l = strtol("-123abz", &endptr, 16);
    printf("strtol(\"-123abz\", &endptr, 16):  %ld  %08X  (\"%s\")\n",
        l, l, endptr);
    l = strtol("123abz", &endptr, 10);
    printf("strtol(\"123abz\", &endptr, 10):   %ld  %08X  (\"%s\")\n",
        l, l, endptr);
    l = strtol("123abz", &endptr, 16);
    printf("strtol(\"123abz\", &endptr, 16):   %ld  %08X  (\"%s\")\n",
        l, l, endptr);
    l = strtol("42 42", &endptr, 13);
    printf("strtol(\"42 42\", &endptr, 13):    %ld  %08X  (\"%s\")\n",
        l, l, endptr);
    nl();

    ul = strtoul("-123abz", &endptr, 10);
    printf("strtoul(\"-123abz\", &endptr, 10): %lu  %08X  (\"%s\")\n",
        ul, ul, endptr);
    ul = strtoul("-123abz", &endptr, 16);
    printf("strtoul(\"-123abz\", &endptr, 16): %lu  %08X  (\"%s\")\n",
        ul, ul, endptr);
    ul = strtoul("123abz", &endptr, 10);
    printf("strtoul(\"123abz\", &endptr, 10):  %lu  %08X  (\"%s\")\n",
        ul, ul, endptr);
    ul = strtoul("123abz", &endptr, 16);
    printf("strtoul(\"123abz\", &endptr, 16):  %lu  %08X  (\"%s\")\n",
        ul, ul, endptr);
    ul = strtoul("42 42", &endptr, 13);
    printf("strtoul(\"42 42\", &endptr, 13):   %lu  %08X  (\"%s\")\n",
        ul, ul, endptr);
    nl();

    d = strtod("-123.456abc", &endptr);
    printf("strtod(\"-123.456abc\", &endptr):  %f %g, %e, (\"%s\")\n",
        d, d, d, endptr);
    d = strtod("-123.456ebc", &endptr);
    printf("strtod(\"-123.456ebc\", &endptr):  %f %g, %e, (\"%s\")\n",
        d, d, d, endptr);
    d = strtod("-123.456e3", &endptr);
    printf("strtod(\"-123.456e3\", &endptr):  %f %g, %e, (\"%s\")\n",
        d, d, d, endptr);
    d = strtod("e", &endptr);
    printf("strtod(\"e\", &endptr):  %f %g, %e, (\"%s\")\n",
        d, d, d, endptr);
    d = strtod("eabc", &endptr);
    printf("strtod(\"eabc\", &endptr):  %f %g, %e, (\"%s\")\n",
        d, d, d, endptr);
    d = strtod("123f", &endptr);
    printf("strtod(\"123f\", &endptr):  %f %g, %e, (\"%s\")\n",
        d, d, d, endptr);
    d = strtod("123.ef", &endptr);
    printf("strtod(\"123.ef\", &endptr):  %f %g, %e, (\"%s\")\n",
        d, d, d, endptr);
    nl();

//- srand(), rand()

    printf("Random values:\n");
    srand(12345);
    for (i = 0; i < 10; i++)
    {
        array[i] = rand() % 100;
        printf("%d ", array[i]);
    }

//- qsort()

    printf("\n\nqsort(array):\n");
    qsort((void *)array, 10, sizeof(int), cmpint);
    for (i = 0; i < 10; i++)
    {
        printf("%d ", array[i]);
    }
    nl();

//- bsearch()

    ptr = (int *)bsearch((const void *)&array[3],
              (const void *)array, 10, sizeof(int), cmpint);
    printf("bsearch(array): %d (%d)\n", *ptr, array[3]);
    nl();

//- system(), getenv()

    system("dir C:\\*.txt");
    nl();

    printf("USERPROFILE=%s\n", getenv("userprofile"));
    nl();

//- mblen()

    setlocale(LC_ALL, "de_DE");

    CharToOem("", ansi);

    printf("mblen(\"%s\")   = %d\n", ansi, mblen("", 10));
    nl();

    CharToOem("This is ansi encoding ", ansi);
    mbstowcs(wide, ansi, sizeof(wide) / sizeof(wchar_t));
    for (i = 0; i < (int)wcslen(wide); i++)
    {
        printf("%04X ", (unsigned short)wide[i]);
    }
    nl();
    printf("ansi: %s\n", ansi);
    wprintf(L"wide: %s\n", wide);
    nl();

    setlocale(LC_ALL, "C");

//- malloc(), calloc(), realloc(), free()

    printf("calloc(), free(), malloc(), realloc():\n");

    ptr = (int *)calloc(100, sizeof(int));
    for (i = 0; i < 100; i++)
        ptr[i] = i;
    for (i = 0; i < 100; i++)
        if (ptr[i] != i)
        {
            error = 1;
            printf("Not equal at %d\n", i);
        }
    free(ptr);

    ptr = (int *)malloc(10 * sizeof(int));
    ptr = (int *)realloc(ptr, 100 * sizeof(int));

    for (i = 0; i < 100; i++)
        ptr[i] = i;
    for (i = 0; i < 100; i++)
        if (ptr[i] != i)
        {
            error = 1;
            printf("Not equal at %d\n", i);
        }
    free(ptr);
    if (!error)
        printf("No errors in alloc functions\n");
    nl();

    /* TODO: abort, atexit, exit, mblen, mbtowc, wctomb,
       calloc, free, malloc, realloc, div, ldiv */

}

// Calls: clearerr, fclose, feof, ferror, fflush, fgetc, fgetpos, fgets
// fopen, fprintf, fputc, fputs, fread, freopen*, fscanf, fseek,
// fsetpos, ftell, fwrite, getc, getchar, gets, perror, printf, putc,
// putchar, puts, remove, rename, rewind, scanf, setbuf*, setvbuf*,
// sprintf, sscanf, tmpfile*, tmpnam, ungetc*, vfprintf, vprintf,
// vsprintf

#define MAX_DBLS 100

typedef double dbls_t[MAX_DBLS];

void testerror(FILE *f, const char *msg)
{
    if (ferror(f))
    {
        printf("Error: %s\n", msg);
        perror("perror");
        nl();
    }
}

char *vout(char *buffer, const char *format, ...)
{
    va_list args;
    FILE *f;
    char buf[MAX_PATH];

//- vfprintf()

    f = fopen("C:\\testfile.txt", "w");
    printf("vfprintf() to C:\\testfile.txt.\n");
    va_start(args, format);
    vfprintf(f, format, args);
    va_end(args);
    fclose(f);

    printf("Reading from C:\\testfile.txt...\n");
    f = fopen("C:\\testfile.txt", "r");
    fgets(buf, sizeof(buf) - 1, f);
    printf("file: %s\n", buf);
    fclose(f);

//- vprintf()

    printf("vprintf(): ");
    va_start(args, format);
    vprintf(format, args);
    va_end(args);

//- vsprintf()

    va_start(args, format);
    vsprintf(buffer, format, args);
    va_end(args);
    return buffer;
}

void stdiotest(void)
{
    char buffer[257];
    char tempname[L_tmpnam + 1];
    char c;
    float flt;
    int i, i2;
    double d;
    dbls_t dbls[2];
    FILE *f;
    fpos_t pos;

    caption("stdio.h");

// -- String I/O ---------------------------------------------------------

//- sprintf(), sscanf()

    sprintf(buffer, "%d %d %f", 100, 20000, log(10));
    printf("String: \"%s\"\n", buffer);
    sscanf(buffer, "%d %d %lf", &i, &i2, &d);
    printf("Values:  %d %d %f\n", i, i2, d);

// -- File I/O -----------------------------------------------------------

    // Create a temporary name first.
    // ------------------------------

//- tmpnam(), perror()

    if (!tmpnam(tempname))
    {
        printf("Could not create temporary name\n");
        exit(1);
    }
    printf("\nTemp name: \"%s\"\n\n", tempname);


    // Test the error functions.
    // -------------------------

    perror("perror test - no error yet");
    nl();

//- fopen(), fprintf(), fclose()

    // First create a little test file.
    f = fopen(tempname, "w");
    fprintf(f, "%d %d %f\n", 10000, 2000, "Hello");
    fprintf(f, "Second line\n");
    fclose(f);

//- fscanf(), ferror(), remove()

    // Then do something wrong, like using fscanf directly after
    // fprintf when opened for read and write.
    f = fopen(tempname, "r+");
    fprintf(f, "%d", rand());
    fscanf(f, buffer, "%s");
    if (ferror(f))
    {
        printf("Error reading from \"%s\"\n", tempname);
        perror("perror");
        nl();
        clearerr(f);
        if (ferror(f))
        {
            printf("Error persists after clearing.\n");
            perror("perror");
            nl();
        }
        else
        {
            printf("Error successfully cleared.\n");
            perror("perror");
            nl();
        }
    }
    else
        printf("Successfully read \"%s\" from file \"%s\".\n\n",
               buffer, tempname);
    fclose(f);
    remove(tempname);


    // Open a file and write some text to it.
    // --------------------------------------

//- fputc(), putc(), fputs(), fflush()

    f = fopen(tempname, "w");
    fprintf(f, "%d %d %f\n", 100, 20000, log(10));
    for (i = 'A'; i <= 'Z'; i++)
    {
        fputc(i, f);
    }
    putc(13, f);
    putc(10, f);
    fputs("abcdefghijklmnopqrstuvwxyz\n\n", f);
    fputs("Msvcrt.dll test program\n", f);
    fprintf(f, "for Delphi unit %s.\n", "MSVCRT.pas");
    fflush(f);
    fclose(f);


    // Open the file and print the text.
    // ---------------------------------

//- fgets()

    f = fopen(tempname, "r");
    printf("File content (%s):\n"
           "----------------------------------------------\n",
           tempname);
    while (fgets(buffer, sizeof(buffer), f))
    {
        printf(buffer);
    }
    fclose(f);
    printf("----------------------------------------------\n\n");


    // Open the file and read values from it.
    // ----------------------------------

    f = fopen(tempname, "r");
    fscanf(f, "%d %d %lf", &i, &i2, &d);
    fscanf(f, "%s", buffer);
    printf("Values in file: %d %d %f\n\"%s\"\n", i, i2, d, buffer);
    fscanf(f, "%s", buffer);
    printf("\"%s\"\n", buffer);
    fscanf(f, "%s", buffer);
    printf("\"%s\"\n\n", buffer);
    fclose(f);


    // Open the file and read it character-wise.
    // -----------------------------------------

//- fgetc()

    f = fopen(tempname, "r");
    printf("Character-wise:\n"
           "----------------------------------------------\n");
    while (!feof(f))
    {
        i = fgetc(f);
        if (i == EOF)
            break;
        printf("%c", i);
    }
    fclose(f);
    printf("----------------------------------------------\n\n");


    // Rename, and try again.
    // ----------------------

//- rename()

    if (rename(tempname, tmpnam(buffer)) == 0)
    {
        strncpy(tempname, buffer, sizeof(tempname));
        printf("Renamed file (%s):\n"
               "----------------------------------------------\n",
               tempname);

        f = fopen(tempname, "r");
        while (fgets(buffer, sizeof(buffer), f))
            printf(buffer);
        fclose(f);
        printf("----------------------------------------------\n\n");
    }
    else
    {
        printf("Could not rename file %s\n\n", tempname);
    }

    // Remove what we dropped.
    remove(tempname);


    // Write and read some binary data.
    // --------------------------------

//- ftell(), fseek(), rewind(), fwrite(), fread(), clearerr()

    // Init first half of array.
    for (i = 0; i < MAX_DBLS; i++)
    {
         dbls[0][i] = 1.1 * i;
    }

    // Write first half to file
    f = fopen(tmpnam(tempname), "wb+");
    fwrite(dbls[0], sizeof(double), MAX_DBLS, f);
    printf("File position:   %ld (%ld)\n", ftell(f), MAX_DBLS * sizeof(double));

    // Rewind and read second half of array from file.
    rewind(f);
    fread(dbls[1], sizeof(double), MAX_DBLS, f);

    // Check first and second half for differences.
    for (i = 0; i < MAX_DBLS; i++)
        if (dbls[0][i] != dbls[1][i])
            printf("Error at %d: %f != %f\n", i, dbls[0][i], dbls[1][i]);

    // Test fseek() and ftell().
    fseek(f, MAX_DBLS * sizeof(double) / 2, SEEK_SET);
    fseek(f, sizeof(double), SEEK_CUR);
    fread(&d, sizeof(d), 1, f);
    testerror(f, "Error after fread");
    printf("Value at offset: %f (%f)\n\n", d, (MAX_DBLS / 2 + 1) * 1.1);

    // Test fgetpos() and fsetpos().
    if (fgetpos(f, &pos))
    {
        testerror(f, "fgetpos()");
        printf("Error in fgetpos\n\n");
        clearerr(f);
    }
    else
    {
        printf("fgetpos(f, &pos): %ld (%ld)\n", pos, (MAX_DBLS / 2 + 2) * sizeof(double));
        fread(&d, sizeof(d), 1, f);
        printf("Read another double: %f\n\n", d);
        if (fsetpos(f, &pos))
        {
            printf("Could not set pos %ld\n", pos);
            perror("perror");
            nl();
        }
        else
        {
            fread(&d, sizeof(d), 1, f);
            printf("After setpos(f, &pos): %f (%f)\n\n", d, (MAX_DBLS / 2 + 2) * 1.1);
        }
    }

    // Clean up the mess.
    fclose(f);
    remove(tempname);

    // Test setvbuf(), setbuf().

    // Test ungetc(), tmpfile().

    // Test vprintf(), vfprintf(), vsprintf().
    printf("vsprintf(): %s\n",
           vout(buffer, "%d %f %g\n", 12345, 123.456, 987.654));

// -- Terminal I/O -------------------------------------------------------

//- gets()

    printf("Enter a string: ");
    gets(buffer);
    for (i = 0; i < (int)strlen(buffer); i++)
        buffer[i] = (char)toupper(buffer[i]);
    printf("String:         %s\n\n", buffer);

//- getchar(), putchar()

    printf("Enter a char:   ");
    c = (char)getchar();
    c = (char)toupper(c);
    printf("Char:           ");
    printf("%c\n\n", c);
    putchar(c);
    putchar(13);
    putchar(10);
    putchar(10);

//- scanf()

    printf("Enter integer float float string.\n");
    printf("Values: ");
    scanf("%d %lf %f %s", &i, &d, &flt, buffer);
    printf("Values: %d %f %f %s\n\n", i, d, flt, buffer);

    /* TODO: freopen, setbuf, setvbuf, tmpfile, ungetc */
}

// Calls: asctime, clock, ctime, difftime, gmtime, localtime, mktime,
// time

void timetest(void)
{
    struct tm *tim;
    time_t tt1, tt2;
    clock_t start, end;
    char s[256];

    caption("time.h");

//- time(), ctime(), localtime(), asctime(), strftime(), gmtime(), mktime(),
//- difftime(), clock()

    tt1 = time(NULL);
    printf("time -> ctime: %s\n\n", ctime(&tt1));

    tim = localtime(&tt1);
    printf("time -> localtime -> asctime: %s\n\n", asctime(tim));

    if (strftime(s, sizeof(s), "%a %b %d %H:%M:%S %Y", tim) > 0)
    {
        printf("strftime: %s\n", s);
    }
    nl();
    
    tim = gmtime(&tt1);
    printf("time -> gmtime -> asctime: %s\n\n", asctime(tim));

    tt2 = mktime(tim);
    printf("%.1f secs?\n\n", difftime(tt1, tt2));

    tim->tm_mon = 1;
    tt1 = mktime(tim);
    printf("tm -> mktime -> ctime: %s\n\n", ctime(&tt1));

    start = clock();
    printf("Please wait a few seconds...\n");
    Sleep(2000);
    end = clock();
    printf("clock: %f\n", (end - start) / CLK_TCK);


}

// Calls: isalnum, isalpha, iscntrl, isdigit, isgraph, isleadbyte,
// islower, isprint, ispunct, isspace, isupper, isxdigit, tolower,
// toupper

static char x(int n)
{
    return ((n == 0) ? (char)'.' : (char)'X');
}

void ctypetest(void)
{
    int c;

    int a1, a2, a3, a4, a5, a6, a7, a8, a9, a10, a11, a12;

    caption("ctype.h");

    for (c = 0; c < 256; c++)
    {
        if (c % 16 == 0)
        {
            printf("  --------------- --- --- --- --- --- --- --- --- --- --- --- ---  --- ---\n");
            printf("  Character:      aln alp ctl dig grp ldb low prn pnc spc upp xdi  low upp\n");
            printf("  --------------- --- --- --- --- --- --- --- --- --- --- --- ---  --- ---\n");
        }

        // This can be done in one function call, but this is easier to debug:
        a1 = isalnum(c);
        a2 = isalpha(c);
        a3 = iscntrl(c);
        a4 = isdigit(c);
        a5 = isgraph(c);
        a6 = isleadbyte(c);
        a7 = islower(c);
        a8 = isprint(c);
        a9 = ispunct(c);
        a10 = isspace(c);
        a11 = isupper(c);
        a12 = isxdigit(c);

        printf("  \'%c\' (%03d, 0x%02X)  %c   %c   %c   %c   %c   %c   %c   %c   %c   %c   %c   %c    %c   %c\n",
               isprint(c) ? c : ' ', c, c,
               x(a1), x(a2), x(a3), x(a4), x(a5), x(a6),
               x(a7), x(a8), x(a9), x(a10), x(a11), x(a12),
               isprint(c) ? tolower(c) : '.',
               isprint(c) ? toupper(c) : '.'
        );
    }
    nl();
}

// Calls: memchr, memcmp, memcpy, memmove, memset, strcat, strchr,
// strcmp, strcoll, strcpy, strcspn, strerror, strftime, strlen,
// strncat, strncmp, strncpy, strpbrk, strrchr, strspn, strstr, strtod,
// strtok, strtol, strtoul, strxfrm

void memcheck(char *m1, char *m2, size_t size)
{
    printf("memcheck(): %d\n\n", memcmp((void *)m1, (void *)m2, size));
}

char *isnull(char *str)
{
    if (str)
        return str;
    else
        return "NULL";
}

void stringtest(void)
{
    char s[256];
    char s2[256];
    int i;
    char *str;

    caption("string.h");

//- memset(), memcmp(), memcpy(), memmove(),memchr()

    memset(s, 0, sizeof(s));
    for (i = 0; i < sizeof(s); i++)
    {
        s[i] = (char)i;
    }
    memset(s2, sizeof(s2), 0);

    printf("memset(), memset() -- s: 1-n; s2: 0-0\n");
    memcheck(s, s2, sizeof(s));

    memcpy(s2, s, sizeof(s));
    printf("memcpy()           -- s: 1-n; s2: 1-n\n");
    memcheck(s, s2, sizeof(s));

    memset(s2 + 4, 11, 10);
    printf("memset(range)      -- s: 1-n; s2: 1-n except 4-15:11\n");
    memcheck(s, s2, sizeof(s));

    memmove(s2 + 4, s + 4, 11);
    printf("memmove(range)     -- s: 1-n; s2: 1-n\n");
    memcheck(s, s2, sizeof(s));

    memset(s2, 'A', sizeof(s2));
    s2[100] = 'B';
    printf("memchr()           -- Position %d in \n\"%s\"\n",
        (char *)memchr(s2, 'B', sizeof(s2)) - (char *)s2,
        s2);
    nl();

//- strerror()

    for (i = 0; i < 20; i++)
    {
        str = strerror(i);
        printf("strerror(%03d):     -- %s\n", i, isnull(str));
    }
    nl();

//- strcat(), strncat(), strcpy(), strncpy(), strcmp(), strncmp(), strlen()

    strcpy(s, "This is the original;");
    printf("strcpy()           -- %s\n", s);
    printf("strlen(s)          -- %d\n", strlen(s));
    strcat(s, " this concatenated to it;");
    printf("strcat()           -- %s\n", s);
    strcat(s, " this too.");
    printf("strcat()           -- %s\n\n", s);
    strncpy(s2, s, sizeof(s2));

    strcpy(s, "0123456789");
    strcpy(s2, "ABCDEFGHIJKLMNOPQRSTUVWXYZ");
    printf("s                  -- %s\n"
           "s2                 -- %s\n", s, s2);
    printf("strncat(s, s2, 15) -- %s\n", strncat(s, s2, 15));
    nl();

    strncpy(s + 5, "-----", 5);
    printf("strncpy()          -- %s\n", s);
    nl();

    strcpy(s2, "0123456789");
    printf("s:                 -- %s\n", s);
    printf("s2:                -- %s\n", s2);
    printf("strcmp(s, s2)      -- %d\n", strcmp(s, s2));
    printf("strncmp(s, s2, 5)  -- %d\n", strncmp(s, s2, 5));
    nl();

//- strtok()

    strcpy(s, "for I := 1 to 10 do\r\nbegin\r\n\tWriteln(I);\r\nend;");
    printf("s:                 -- \"\n\%s\n\"\n", s);
    str = strtok(s, TOK_SEPARATORS);
    while (str != NULL)
    {
        printf("strtok()           -- \"%s\"\n", str);
        str = strtok(NULL, TOK_SEPARATORS);
    }
    nl();

//- setlocale(), strcoll(), strxfrm()

    setlocale(LC_ALL, "de_DE");
    printf("Locale \"de_DE\":\n");

    CharToOem("Strae", s);
    CharToOem("Strasse", s2);
    printf("s:                 -- %s\n", s);
    printf("s2:                -- %s\n", s2);
    printf("strcoll(s, s2)     -- %d\n", strcoll("Strae", "Strasse"));
    nl();

    printf("strxfrm(\"%s\")  -- ", s);
    strxfrm(s, "Strae", sizeof(s));
    printf("%s\n", s);
    printf("strxfrm(\"%s\") -- ", s2);
    strxfrm(s2, "Strasse", sizeof(s2));
    printf("%s\n", s2);
    printf("strcmp(s, s2)      -- %d\n", strcmp(s, s2));
    nl();

    setlocale(LC_ALL, "C");
    printf("Locale \"C\":\n");

    CharToOem("Strae", s);
    CharToOem("Strasse", s2);
    printf("s:                 -- %s\n", s);
    printf("s2:                -- %s\n", s2);
    printf("strcoll(s, s2)     -- %d\n", strcoll("Strae", "Strasse"));
    nl();

    printf("strxfrm(\"%s\")  -- ", s);
    strxfrm(s, "Strae", sizeof(s));
    printf("%s\n", s);
    printf("strxfrm(\"%s\") -- ", s2);
    strxfrm(s2, "Strasse", sizeof(s2));
    printf("%s\n", s2);
    printf("strcmp(s, s2)      -- %d\n", strcmp(s, s2));
    nl();

// strchr(), strrchr(), strpbrk()

    printf("s:                 -- %s\n", strcpy(s, "Abcdefggfedcba"));
    printf("strchr(s, \'c\'):    -- %s\n", isnull(strchr(s, 'c')));
    printf("strchr(s, \'g\'):    -- %s\n", isnull(strchr(s, 'g')));
    printf("strchr(s, \'z\'):    -- %s\n", isnull(strchr(s, 'z')));
    nl();

    printf("strrchr(s, \'c\'):   -- %s\n", isnull(strrchr(s, 'c')));
    printf("strrchr(s, \'g\'):   -- %s\n", isnull(strrchr(s, 'g')));
    printf("strrchr(s, \'z\'):   -- %s\n", isnull(strrchr(s, 'z')));
    nl();

    printf("strpbrk(s, \"ghi\")  -- %s\n", isnull(strpbrk(s, "ghi")));
    printf("strpbrk(s, \"xyz\")  -- %s\n", isnull(strpbrk(s, "xyz")));
    printf("strpbrk(s, \"Abc\")  -- %s\n", isnull(strpbrk(s, "Abc")));
    printf("strpbrk(s, \"ayz\")  -- %s\n", isnull(strpbrk(s, "ayz")));
    nl();

//- strspn(), strcspn(), strpbrk()

    printf("strspn(s, \"ghi\")   -- %u\n", strspn(s, "ghi"));
    printf("strspn(s, \"xyz\")   -- %u\n", strspn(s, "xyz"));
    printf("strspn(s, \"Abc\")   -- %u\n", strspn(s, "Abc"));
    printf("strspn(s, \"ayz\")   -- %u\n", strspn(s, "ayz"));
    nl();

    printf("strcspn(s, \"ghi\")  -- %u\n", strcspn(s, "ghi"));
    printf("strcspn(s, \"xyz\")  -- %u\n", strcspn(s, "xyz"));
    printf("strcspn(s, \"Abc\")  -- %u\n", strcspn(s, "Abc"));
    printf("strcspn(s, \"ayz\")  -- %u\n", strcspn(s, "ayz"));
    nl();

//- strstr()

    printf("strstr(s, \"ghi\")   -- %s\n", isnull(strstr(s, "ghi")));
    printf("strstr(s, \"xyz\")   -- %s\n", isnull(strstr(s, "xyz")));
    printf("strstr(s, \"Abc\")   -- %s\n", isnull(strstr(s, "Abc")));
    printf("strstr(s, \"fgg\")   -- %s\n", isnull(strstr(s, "fgg")));
    nl();
}

// Calls: fgetwc, fputwc, fwprintf, fwscanf, swprintf, swscanf,
// ungetwc, vfwprintf, vswprintf, vwprintf, wcscat, wcschr, wcscmp,
// wcscoll, wcscpy, wcscspn, wcsftime, wcslen, wcsncat, wcsncmp,
// wcsncpy, wcspbrk, wcsrchr, wcsspn, wcsstr, wcstod, wcstok, wcstol,
// wcstombs, wcstoul, wcsxfrm, wprintf, fgetws, fputws, getwc, getwchar,
// putwc, putwchar

void wchartest(void)
{
    caption("wchar.h");
}




